@@ -5,7 +5,7 @@ import scipy.stats as st
 import os, subprocess
 
 from .extract_prs import check_pfiles
-from .tools import get_plink_path
+from .tools import get_plink_path, run_plink_command
 
 
 def association_test_func_plink2(data, covar_list, standardize, name, data_pheno, pheno_type):
@@ -97,13 +97,7 @@ def _run_plink2_assoc_test(
         if "scales vary too widely" in str(e):
             print("The association test failed because of numerical instability in the covariates. Rescaling the covariates.")
             command.extend(["--covar-variance-standardize"])
-            try:
-                subprocess.run(command, capture_output=True, text=True, check=True)
-            except Exception as e:
-                print(f"Error running PLINK command: {e}")
-                print(f"PLINK stdout: {e.stdout}")
-                print(f"PLINK stderr: {e.stderr}")
-                raise ValueError("PLINK command failed. Check the error messages above for details.")
+            run_plink_command(command)
             
         else:
             print(f"Error running PLINK command: {e}")
@@ -162,6 +156,12 @@ def _prepare_psam_file(genetic_path, data_pheno, pheno_type, standardize):
     # Read the PSAM file
     psam = pd.read_csv(genetic_path + ".psam", delimiter="\t")
     
+    # Ensure IID column types match before merging to prevent errors
+    if '#IID' in psam.columns and 'IID' in data_pheno.columns:
+        psam['#IID'] = psam['#IID'].astype(data_pheno['IID'].dtype)
+    elif 'IID' in psam.columns and 'IID' in data_pheno.columns:
+         psam['IID'] = psam['IID'].astype(data_pheno['IID'].dtype)
+
     # Merge phenotype data with the PSAM dataframe depending on column present
     if "#FID" in psam.columns:
         data_pheno_trait = data_pheno[["FID", "IID", "PHENO"]].rename(columns={"FID": "#FID", "PHENO": "PHENO1"}).copy()
