@@ -1,15 +1,12 @@
 import pandas as pd
 import numpy as np
 import scipy.stats as st
-import os, subprocess
-import shutil
+import os
 import warnings
 from collections import Counter
-import wget
-
-from .constants import STANDARD_COLUMNS, BUCKET_URL
-from .tools import read_config
 
+from .constants import STANDARD_COLUMNS
+from .tools import get_reference_panel_path, run_plink_command, create_tmp, get_plink_path
 
 
 def remove_na(data):
@@ -282,7 +279,7 @@ def fill_snpids_func(data, reference_panel_df, keep_indel):
     return data
 
 def check_int_column(data, int_col):
-    """Set the type of the int_col column to Int64 and non-numeric values to NA."""
+    """Set the type of the int_col column to Int64 and non-numeric values to NA. This function is used to check the validity of the CHR and POS columns."""
     nrows = data.shape[0]
     # Remove any non-digit characters, convert to numeric, setting non-numeric to NaN
     data[int_col] = pd.to_numeric(data[int_col].astype(str).str.extract('(\d+)', expand=False), errors='coerce')
@@ -403,53 +400,39 @@ def check_arguments(
 
 def save_data(data, name, path="", fmt="h5", sep="\t", header=True):
     """
-    Save a DataFrame to a file in a given format.
+    Save data to a specified file format.
+
+    Supported formats: .h5 (default), .csv, .txt.
+    Future supported formats: .vcf, .vcf.gz.
 
     Args:
-    - data (pd.DataFrame): The data to be saved.
-    - name (str): The name of the file without extension.
-    - path (str, optional): Directory path for saving. Default is the current directory.
-    - fmt (str, optional): Format for the file, e.g., "h5", "csv", "txt". Default is "h5".
-    - sep (str, optional): Delimiter for csv or txt files. Default is tab.
-    - header (bool, optional): Whether to include header in csv or txt files. Default is True.
-
-    Returns:
-    None. But saves the data to a file and prints the file path.
-
-    Raises:
-    - ValueError: If the provided format is not recognized.
+        data (pd.DataFrame): DataFrame to be saved.
+        name (str): A unique identifier for the data, used as the filename.
+        path (str, optional): The directory where the file will be saved. Defaults to current directory.
+        fmt (str, optional): The desired file format. Defaults to "h5".
+        sep (str, optional): Delimiter for text-based formats (.csv, .txt). Defaults to tab.
+        header (bool, optional): Whether to include column names in text-based formats. Defaults to True.
     """
-    if path:
-        path_name = os.path.join(path, f"{name}.{fmt}")
-    else:
-        path_name = f"{name}.{fmt}"
-
+    path = os.path.join(path, name)
     if fmt == "h5":
-        df = data.copy()
-        for col in df.select_dtypes(include="integer").columns:
-            df[col] = df[col].astype("float64")
-        df.to_hdf(path_name, mode="w", key="data")
-
-    elif fmt in ["csv", "txt"]:
-        data.to_csv(path_name, sep=sep, header=header, index=False)
-
+        data.to_hdf(f"{path}.h5", key="data", mode="w", format="table")
+    elif fmt == "csv":
+        data.to_csv(f"{path}.csv", sep=sep, header=header, index=False)
+    elif fmt == "txt":
+        data.to_csv(f"{path}.txt", sep=sep, header=header, index=False)
     else:
-        raise ValueError(
-            "The fmt argument takes value in (h5 (default), csv, txt)."
-        )
-
-    print(f"Data saved to {path_name}")
+        print(f"Format {fmt} is not supported yet.")
 
 
 def Combine_Geno(Gs):
     """
-    Combine a list of GWAS objects into one.
+    Combine multiple Geno instances.
 
     Args:
-    - Gs (list): List of GWAS objects.
+        Gs (list): A list of Geno instances to combine.
 
     Returns:
-    Geno object: Combined Geno object.
+        Geno: A new Geno instance containing the combined data.
     """
     from .Geno import Geno
     
@@ -462,118 +445,101 @@ def Combine_Geno(Gs):
 
     return Geno(C)
 
-def filter_by_gene_func(data, gene_identifier, id_type="symbol", window_size=1000000, build="37"):
+
+def update_eaf_func(data, reference_panel, object_name, ram=10000, fill=True):
     """
-    Filtering the data to include only variants that are within a specified distance of a specific gene.
-    Corresponds to the :meth:`Geno.filter_by_gene` method.
-    Args:
-        data (pd.DataFrame): Input data with at least 'CHR' and 'POS' columns.
-        gene_identifier (str): Identifier for the gene/protein to filter variants around.
-        id_type (str, optional): Type of identifier provided. Options are:
-            - "symbol": Gene symbol (e.g., "APOE")
-            - "HGNC": HGNC ID (e.g., "HGNC:613")
-            - "name": Full gene name (e.g., "apolipoprotein E")
-            - "Ensembl": Ensembl gene ID (e.g., "ENSG00000130203")
-            - "NCBI": NCBI gene ID (e.g., "348")
-            - "UCSC": UCSC gene ID (e.g., "uc001hbu.2")
-            - "Vega": Vega gene ID (e.g., "OTTHUMG00000019505")
-            Default is "symbol".
-        window_size (int, optional): Size of the window around the gene in base pairs. Default is 1,000,000 (1Mb).
-        build (str, optional): Genome build of the data. Default is "37".
-        
-    Returns:
-        pd.DataFrame: Filtered DataFrame containing only variants within the specified window 
-            around the gene, with additional column 'Distance'.
+    Core logic to update or create the EAF (Effect Allele Frequency) column.
 
-    Notes:
-        - Distance is calculated from the nearest gene boundary (start or end position)
-        - Null distances indicate the variant is within the gene
+    This function calculates EAF from a reference panel. If CHR/POS are available,
+    it uses a fast, coordinate-based extraction with PLINK. Otherwise, it falls
+    back to SNP-ID-based extraction.
     """
+    ref_panel_path, ref_filetype = get_reference_panel_path(reference_panel)
+    create_tmp()
+
+    by_coordinate = "CHR" in data.columns and "POS" in data.columns
+
+    # --- Match by CHR/POS or SNP ID ---
+    if by_coordinate:
+        print("CHR/POS columns present. SNPs searched based on genomic positions.")
         
-    # Validate id_type
-    valid_id_types = ["symbol", "HGNC_id", "name", "gene_id", "NCBI_id", "UCSC_id", "Vega_id"]
-    if id_type in ["HGNC", "NCBI", "UCSC", "Vega"]:
-        id_type = id_type + "_id"
-    if id_type == "Ensembl":
-        id_type = "gene_id"
-    if id_type not in valid_id_types:
-        raise ValueError(f"Invalid id_type. Must be one of: {', '.join(valid_id_types)}")
-    
-    # Validate build
-    if int(build) not in [37, 38]:
-        raise ValueError(f"Invalid build. Must be one of: 37, 38")
-    
-    # Download the gene info file if not already present in the reference folder
-    config = read_config()
-    ref_path = config["paths"]["ref_path"]
-    gene_info_file = os.path.join(ref_path, "gene_id_mapping_filtered.parquet")
-    if not os.path.exists(gene_info_file):
-        # Download parquet file
-        print(f"Downloading gene info file to {gene_info_file}...")    
-        url = BUCKET_URL + "gene_id_mapping_filtered.parquet"
-        try:
-            wget.download(url, gene_info_file)
-            print("\nDownload complete.")
-        except Exception as e:
-            if os.path.exists(gene_info_file):
-                os.remove(gene_info_file)
-            raise RuntimeError(f"Failed to download gene info: {e}")
+        # 1. Write coordinates to a temp file for PLINK's --extract range
+        coord_path = os.path.join("tmp_GENAL", f"{object_name}_coord_list.txt")
+        data[['CHR', 'POS', 'POS']].dropna().to_csv(coord_path, sep='\t', index=False, header=False)
 
-    df_gene_info = pd.read_parquet(gene_info_file, engine="pyarrow")
-    
-    # Find gene coordinates
-    gene_data = df_gene_info[df_gene_info[id_type] == gene_identifier]
-    
-    if gene_data.empty:
-        raise ValueError(f"Gene with {id_type}='{gene_identifier}' not found in gene info database.")
-    
-    if len(gene_data) > 1:
-        print(f"Warning: Multiple entries found for {id_type}='{gene_identifier}'. Using the first entry.")
-    gene_data = gene_data.iloc[0,:]
+        # 2. Run --freq directly, extracting by range and adding POS to output
+        freq_prefix = os.path.join("tmp_GENAL", f"{object_name}_eaf_freqs")
+        plink_command = (
+            f"{get_plink_path()} --{'pfile' if ref_filetype == 'pgen' else 'bfile'} {ref_panel_path} "
+            f"--memory {ram} "
+            f"--extract range {coord_path} "
+            f"--freq cols=+pos "
+            f"--out {freq_prefix}"
+        )
+        run_plink_command(plink_command)
 
-    print(f"Filtering variants within {window_size}bp window based on genome build {build} around gene: {', '.join(f'{col}: {gene_data[col]}' for col in valid_id_types)}")
-    
-    # Extract gene location information
-    chrom = gene_data['CHR']
-    # Convert to integer if possible
-    if str(chrom).isdigit():
-        chrom = int(chrom)
-    elif chrom=="X":
-        chrom=23
+        # 3. Load frequency results
+        freq_path = f"{freq_prefix}.afreq"
+        if not os.path.exists(freq_path) or os.path.getsize(freq_path) == 0:
+            warnings.warn("No variants from your data were found in the reference panel by coordinate, or PLINK failed.")
+            return data.copy()
+        
+        freqs_df = pd.read_csv(freq_path, sep='\t')
+        freqs_df.rename(columns={'#CHROM': 'CHR', 'ALT': 'ALT_calc', 'ALT_FREQS': 'EAF_ref'}, inplace=True)
+        
+        # 4. Merge with original data to get EA and compute final EAF
+        data = data.merge(freqs_df[["CHR", "POS", "ALT_calc", "EAF_ref"]], on=['CHR', 'POS'], how='left')
+        
     else:
-        raise ValueError(f"Gene {gene_identifier} is located on chromosome {chrom}, which is not supported.")
-    
-    gene_start = int(gene_data[f'gene_start_{build}'])
-    gene_end = int(gene_data[f'gene_end_{build}'])
+        print("Using SNP IDs to extract frequencies.")
+        if "SNP" not in data.columns:
+            raise ValueError("SNP column is required when CHR/POS are not available.")
+                    
+        snp_list_path = os.path.join("tmp_GENAL", f"{object_name}_snp_list.txt")
+        data[["SNP"]].dropna().to_csv(snp_list_path, index=False, header=False)
 
-    # Define the window boundaries
-    window_start = max(0, gene_start - window_size/2)
-    window_end = gene_end + window_size/2
-    
-    # Filter variants within the window
-    filtered = data[
-        (data['CHR'] == chrom) & 
-        (data['POS'] >= window_start) & 
-        (data['POS'] <= window_end)
-    ].copy()
+        freq_prefix = os.path.join("tmp_GENAL", f"{object_name}_eaf_freqs")
+        plink_command = (
+            f"{get_plink_path()} --{'pfile' if ref_filetype == 'pgen' else 'bfile'} {ref_panel_path} "
+            f"--memory {ram} "
+            f"--extract {snp_list_path} "
+            f"--freq "
+            f"--out {freq_prefix}"
+        )
+        run_plink_command(plink_command)
 
-    if not filtered.empty:
-        # Calculate distance from gene: if inside the gene, distance is 0, if before, distance is negative, if after, distance is positive
-        filtered.loc[:, 'Distance'] = np.nan
-        
-        # Create boolean masks
-        mask_inside = filtered['POS'].between(gene_start, gene_end)
-        mask_before = filtered['POS'] < gene_start
-        mask_after  = filtered['POS'] > gene_end
+        freq_path = f"{freq_prefix}.afreq"
+        if not os.path.exists(freq_path):
+            warnings.warn("PLINK did not generate a frequency file. Cannot update EAF.")
+            return data.copy()
 
-        filtered.loc[mask_inside, 'Distance'] = 0
-        filtered.loc[mask_before, 'Distance'] = filtered['POS'] - gene_start
-        filtered.loc[mask_after, 'Distance']  = filtered['POS'] - gene_end
+        freqs_df = pd.read_csv(freq_path, sep='\t')
+        freqs_df.rename(columns={"#CHROM": "CHR", "ID": "SNP", "ALT": "ALT_calc", "ALT_FREQS": "EAF_ref"}, inplace=True)
 
-        filtered["Distance"] = filtered["Distance"].astype("Int64")
-        
-        print(f"Found {len(filtered)} variants.")
+        data = data.merge(freqs_df[["SNP", "ALT_calc", "EAF_ref"]], on="SNP", how="left")
+
+    if data["EAF_ref"].isna().all():
+        warnings.warn("No matching SNPs found in the reference panel.")
+        return data.copy()
+
+    # Handle allele direction to ensure correct EAF is returned
+    conditions = [
+        data["EA"] == data["ALT_calc"],
+        data["NEA"] == data["ALT_calc"],
+    ]
+    choices = [
+        data["EAF_ref"],
+        1 - data["EAF_ref"],
+    ]
+    data["EAF_new"] = np.select(conditions, choices, default=np.nan)
+
+    # Create updated EAF column
+    if 'EAF' not in data.columns:
+        data['EAF'] = np.nan
+    if fill:
+        data['EAF'] = np.where(pd.notna(data["EAF_new"]), data["EAF_new"], data['EAF'])
     else:
-        print(f"No variants found in a {window_size}bp window around {gene_identifier}")
-    
-    return filtered
\ No newline at end of file
+        data['EAF'] = np.where(pd.notna(data["EAF_new"]), data["EAF_new"], np.nan)
+    data.drop(columns=["EAF_new", "EAF_ref", "ALT_calc"], inplace=True)
+
+    return data
\ No newline at end of file
